// NOTE: This file can only require from the standard library, as it is ran as a
// separate node process

// Before enabling, keep in mind that some processes (e.g. git when spawning gpg)
// process stderr as well as stdout, so logging to stderr can cause failures
const debugLogEnabled = process.env.GITKRAKEN_DEBUG_ASKPASS;
const debugLogToProcessStderr = (message) => {
  if (debugLogEnabled) {
    process.stderr.write(message);
  }
};

debugLogToProcessStderr(`IN ASKPASS, args: ${process.argv.join(',')}\n`);

const net = require('net');

// Make sure to keep in sync with `AskPassConstants.js`
const askPassFunctions = {
  GIT_ASKPASS: 'gitAskpass'
};

const {
  GITKRAKEN_ASKPASS_FUNCTION = askPassFunctions.GIT_ASKPASS,
  GITKRAKEN_ASKPASS_SESSION_ID,
  GITKRAKEN_SOCKET_SERVICE_PORT
} = process.env;

const [, , ppid, ...args] = process.argv;

const askPass = () => {
  const request = {
    s: GITKRAKEN_ASKPASS_SESSION_ID,
    p: parseInt(ppid, 10),
    t: GITKRAKEN_ASKPASS_FUNCTION,
    f: null, // request field
    u: null // url for request
  };

  const promptArg = args.join(' ');
  const usernameMatch = /^Username for (.+)/g.exec(promptArg);
  const passwordMatch = /^Password for (.+)/g.exec(promptArg);
  const passphraseMatch = /^Enter passphrase for key (.+)/g.exec(promptArg);

  let match;
  if (usernameMatch) {
    request.f = 'u';
    match = usernameMatch;
  } else if (passwordMatch) {
    request.f = 'pw';
    match = passwordMatch;
  } else if (passphraseMatch) {
    request.f = 'p';
    match = passphraseMatch;
  }

  if (!request.f) {
    process.exit(1);
  }

  let url = match[1].trim();
  if (url.endsWith(':')) {
    url = url.substr(0, url.length - 1);
  }

  // Handle single quoted URLs (SSH)
  const urlMatch = /^'(.+)'$/.exec(url);
  if (urlMatch?.[1]) {
    request.u = urlMatch[1];
  } else {
    request.u = url;
  }

  setupClient((data) => {
    const {
      r: result,
      c: exitCode
    } = JSON.parse(data.toString());
    if (result) {
      process.stdout.write(result);
      process.exit(0);
    } else if (exitCode) {
      process.exit(exitCode);
    } else {
      process.exit(1);
    }
  }, JSON.stringify(request));
};

const performHandshake = clientSocket => new Promise((resolve, reject) => {
  clientSocket.on('error', reject);
  clientSocket.on('close', reject);
  clientSocket.on('end', reject);

  debugLogToProcessStderr('waiting for handshake\n');
  clientSocket.once('data', (data) => {
    debugLogToProcessStderr(`handshake data: ${data.toString()}\n`);
    clientSocket.removeListener('error', reject);
    clientSocket.removeListener('close', reject);
    clientSocket.removeListener('end', reject);

    if (data.toString() !== 'ready') {
      reject();
    } else {
      resolve();
    }
  });

  clientSocket.write('ask_pass');
});

// a NUL byte (\0) will be appended to dataToWrite
const setupClient = (onData, dataToWrite, exitOnEnd = true) => {
  const client = net.createConnection(
    {
      port: GITKRAKEN_SOCKET_SERVICE_PORT,
      host: 'localhost'
    },
    async () => {
      try {
        await performHandshake(client);
      } catch (error) {
        debugLogToProcessStderr(`handshake error: ${error.message}\n`);
        process.exit(1);
      }

      client.on('error', () => {
        debugLogToProcessStderr('client error\n');
        process.exit(1);
      });
      client.on('end', () => {
        debugLogToProcessStderr('client end\n');
        if (exitOnEnd) {
          process.exit(1);
        }
      });

      client.on('data', (data) => {
        try {
          debugLogToProcessStderr(`client data(${data.length}): ${data.toString()}`);
          onData(data);
        } catch (error) {
          debugLogToProcessStderr(`client data error: ${error.message}\n`);
          process.exit(1);
        }
      });
      client.write(`${dataToWrite}\0`);
      debugLogToProcessStderr(`wrote data: ${dataToWrite} + NUL`);
    }
  );
};

debugLogToProcessStderr(`GITKRAKEN_ASKPASS_FUNCTION: ${GITKRAKEN_ASKPASS_FUNCTION}`);
switch (GITKRAKEN_ASKPASS_FUNCTION) {
  case askPassFunctions.GIT_ASKPASS:
    askPass();
    break;

  default:
    process.exit(1);
}
